import sys

# get point of interest
lat_deg = sys.argv[1]
lat_min = sys.argv[2]
lat_hem = sys.argv[3]
lon_deg = sys.argv[4]
lon_min = sys.argv[5]
lon_hem = sys.argv[6]

# get extent of bounding box
extent_km = 100.0 / 2.0
if len(sys.argv) > 7:
    extent_km = float(sys.argv[7]) / 2.0
extent = extent_km * 90 / 10000 # convert to degrees

# remove special characters
if lat_deg[-1] == '°':
    lat_deg = lat_deg[:-1]
if lat_min[-1] == "'":
    lat_min = lat_min[:-1]
if lon_deg[-1] == '°':
    lon_deg = lon_deg[:-1]
if lon_min[-1] == "'":
    lon_min = lon_min[:-1]

# convert to numbers
lat_deg = float(lat_deg)
lat_min = float(lat_min)
lon_deg = float(lon_deg)
lon_min = float(lon_min)

# convert to decimal
lat = lat_deg + (lat_min / 60)
if lat_hem == 'S':
    lat *= -1
elif lat_hem != 'N':
    raise RuntimeError(f'Invalid hemisphere designation (only N or S allowed)')
lon = lon_deg + (lon_min / 60)
if lon_hem == 'W':
    lon *= -1
elif lon_hem != 'E':
    raise RuntimeError(f'Invalid hemisphere designation (only E or W allowed)')

# build bounding box
l_lon = lon - extent
if l_lon < -180.0:
    l_lon = 360.0 + l_lon
r_lon = lon + extent
if r_lon > 180.0:
    r_lon = -360.0 + r_lon
u_lat = lat + extent
if u_lat > 90.0:
    u_lat = 180.0 - u_lat
l_lat = lat - extent
if l_lat < -90.0:
    l_lat = -180.0 + l_lat

# build coordinate strings
coord_strs = [
    f'[ {l_lon}, {u_lat} ]', # upper left
    f'[ {l_lon}, {l_lat} ]', # lower left
    f'[ {r_lon}, {l_lat} ]', # lower right
    f'[ {r_lon}, {u_lat} ]', # upper right
    f'[ {l_lon}, {u_lat} ]'  # upper left
]

# build name string
name_str = f'{int(lat_deg)}_{int(lat_min)}_{lat_hem}__{int(lon_deg)}_{int(lon_min)}_{lon_hem}'

# build geojson string
geojson_str = """{
  "type": "FeatureCollection",
  "name": "$NAME",
  "crs": {
    "type": "name",
    "properties": { "name": "urn:ogc:def:crs:OGC:1.3:CRS84" }
  },
  "features": [ {
    "type": "Feature",
    "properties": { },
    "geometry": { "type": "Polygon", "coordinates": [ [ $COORDS ] ] }
  } ]
}""".replace("$COORDS", ", ".join(coord_strs)).replace("$NAME", name_str)

# return geojson str
print(geojson_str)



#
# This script sourced from LPDAAC git repository: https://git.earthdata.nasa.gov/projects/LPDUR/repos/daac_data_download_python
#

"""
---------------------------------------------------------------------------------------------------
 How to Access the LP DAAC Data Pool with Python
 The following Python code example demonstrates how to configure a connection to download data from
 an Earthdata Login enabled server, specifically the LP DAAC's Data Pool.
---------------------------------------------------------------------------------------------------
 Original Author: Cole Krehbiel
 Date of Version Sourced: 05/14/2020
---------------------------------------------------------------------------------------------------
"""
# Load necessary packages into Python
from netrc import netrc
import argparse
import datetime
import os
import requests

# ----------------------------------USER-DEFINED VARIABLES--------------------------------------- #
# Set up command line arguments
parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('-dir', '--directory', required=True, help='Specify directory to save files to')
parser.add_argument('-f', '--files', required=True, help='A single granule URL, or the location of csv or textfile containing granule URLs')
args = parser.parse_args()

saveDir = args.directory  # Set local directory to download to
files = args.files        # Define file(s) to download from the LP DAAC Data Pool

# ---------------------------------SET UP WORKSPACE---------------------------------------------- #
# Create a list of files to download based on input type of files above
if files.endswith('.txt') or files.endswith('.csv'):
    fileList = open(files, 'r').readlines()  # If input is textfile w file URLs
elif isinstance(files, str):
    fileList = [files]                       # If input is a single file

# Generalize download directory
if saveDir[-1] != '/' and saveDir[-1] != '\\':
    saveDir = saveDir.strip("'").strip('"') + os.sep
    if not os.path.exists(saveDir):
        os.makedirs(saveDir)

# --------------------------------AUTHENTICATION CONFIGURATION----------------------------------- #
# Determine if netrc file exists, and if so, if it includes NASA Earthdata Login Credentials
urs = 'urs.earthdata.nasa.gov'    # Address to call for authentication
netrcDir = os.path.expanduser("~/.netrc")
netrc(netrcDir).authenticators(urs)[0]
netrc(netrcDir).authenticators(urs)[2]

# -----------------------------------------DOWNLOAD FILE(S)-------------------------------------- #
# Loop through and download all files to the directory specified above, and keeping same filenames
for f in fileList:
    file_name = os.path.join(saveDir, f.split('/')[-1].strip())
    file_date = file_name.split('_')[2]
    file_year = file_date[:4]
    file_doy = file_date[4:7]
    file_date_time = datetime.datetime.strptime('{}-{}'.format(file_year, file_doy),'%Y-%j')
    file_month = '{}'.format(file_date_time.month).rjust(2,'0')
    file_day = '{}'.format(file_date_time.day).rjust(2,'0')
    file_version = file_name.split('_')[9].split('.')[0][1:]
    file_shortname = file_name[2:10]
    file_url = 'https://e4ftl01.cr.usgs.gov/GEDI/{}.{}/{}.{}.{}/{}'.format(file_shortname, file_version, file_year, file_month, file_day, file_name[2:])
    # Create and submit request and download file
    with requests.get(file_url, verify=False, stream=True, auth=(netrc(netrcDir).authenticators(urs)[0], netrc(netrcDir).authenticators(urs)[2])) as response:
        if response.status_code != 200:
            print("Error downloading {} - {} {}".format(file_url, response, response.content))
        else:
            response.raw.decode_content = True
            content = response.raw
            with open(file_name, 'wb') as d:
                while True:
                    chunk = content.read(16 * 1024)
                    if not chunk:
                        break
                    d.write(chunk)
            print('Downloaded file: {}'.format(file_name))



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import netrc
import getpass
import requests
import threading
import socket
import json
import struct
import ctypes
import time
import logging
import numpy
from sliderule import version

###############################################################################
# LOGGING
###############################################################################

logger = logging.getLogger(__name__)
eventlogger = {
    0: logger.debug,
    1: logger.info,
    2: logger.warning,
    3: logger.error,
    4: logger.critical
}

###############################################################################
# DNS OVERRIDE
###############################################################################

sliderule_getaddrinfo = socket.getaddrinfo
sliderule_dns = {}

def __override_getaddrinfo(*args, **kwargs):
    global sliderule_getaddrinfo, sliderule_dns
    url = args[0].lower()
    if url in sliderule_dns:
        logger.debug("getaddrinfo returned {} for {}".format(sliderule_dns[url], url))
        return sliderule_getaddrinfo(sliderule_dns[url], *args[1:], **kwargs)
    else:
        return sliderule_getaddrinfo(*args, **kwargs)

socket.getaddrinfo = __override_getaddrinfo

###############################################################################
# GLOBALS
###############################################################################

EXCEPTION_CODES = {
    "INFO": 0,
    "ERROR": -1,
    "TIMEOUT": -2,
    "RESOURCE_DOES_NOT_EXIST": -3,
    "EMPTY_SUBSET": -4,
    "SIMPLIFY": -5
}

BASIC_TYPES = {
    "INT8":     { "fmt": 'b', "size": 1, "nptype": numpy.int8   },
    "INT16":    { "fmt": 'h', "size": 2, "nptype": numpy.int16  },
    "INT32":    { "fmt": 'i', "size": 4, "nptype": numpy.int32  },
    "INT64":    { "fmt": 'q', "size": 8, "nptype": numpy.int64  },
    "UINT8":    { "fmt": 'B', "size": 1, "nptype": numpy.uint8  },
    "UINT16":   { "fmt": 'H', "size": 2, "nptype": numpy.uint16 },
    "UINT32":   { "fmt": 'I', "size": 4, "nptype": numpy.uint32 },
    "UINT64":   { "fmt": 'Q', "size": 8, "nptype": numpy.uint64 },
    "BITFIELD": { "fmt": 'x', "size": 0, "nptype": numpy.byte   }, # unsupported
    "FLOAT":    { "fmt": 'f', "size": 4, "nptype": numpy.single },
    "DOUBLE":   { "fmt": 'd', "size": 8, "nptype": numpy.double },
    "TIME8":    { "fmt": 'q', "size": 8, "nptype": numpy.int64  }, # numpy.datetime64
    "STRING":   { "fmt": 's', "size": 1, "nptype": numpy.byte   }
}

CODED_TYPE = {
    0:  "INT8",
    1:  "INT16",
    2:  "INT32",
    3:  "INT64",
    4:  "UINT8",
    5:  "UINT16",
    6:  "UINT32",
    7:  "UINT64",
    8:  "BITFIELD",
    9:  "FLOAT",
    10: "DOUBLE",
    11: "TIME8",
    12: "STRING"
}

###############################################################################
# CLASSES
###############################################################################

#
# FatalError
#
class FatalError(RuntimeError):
    pass

#
# RetryRequest
#
class RetryRequest(RuntimeError):
    pass

#
# Session
#
class Session:

    PUBLIC_URL = "slideruleearth.io"
    PUBLIC_ORG = "sliderule"
    MAX_PS_CLUSTER_WAIT_SECS = 600

    #
    # constructor
    #
    def __init__ (self,
        domain          = PUBLIC_URL,
        verbose         = False,
        loglevel        = logging.INFO,
        organization    = 0,
        desired_nodes   = None,
        time_to_live    = 60, # minutes
        bypass_dns      = False,
        trust_env       = False,
        ssl_verify      = True,
        rqst_timeout    = (10, 120), # (connection, read) in seconds
        log_handler     = None,
        decode_aux      = True,
        rethrow         = False ):
        '''
        creates and configures a sliderule session
        '''
        # configure parameter arguments
        self.session = requests.Session()
        self.session.trust_env = trust_env
        self.ssl_verify = ssl_verify
        self.rqst_timeout = rqst_timeout
        self.decode_aux = decode_aux
        self.throw_exceptions = rethrow

        # initialize to empty
        self.ps_refresh_token = None
        self.ps_access_token = None
        self.ps_token_exp = None
        self.ps_lock = threading.Lock()
        self.console = None
        self.recdef_table = {}
        self.arrow_file_table = {} # for processing arrowrec records
        self.local_dns = {} # global dictionary of DNS entries to override

        # configure logging
        self.set_verbose(verbose, loglevel)
        if log_handler != None:
            logger.addHandler(log_handler)

        # configure domain
        self.service_domain = domain

        # configure callbacks
        self.callbacks = {
            'eventrec': Session.__logrec,
            'exceptrec': Session.__alertrec,
            'arrowrec.meta': Session.__arrowrec,
            'arrowrec.data': Session.__arrowrec,
            'arrowrec.eof': Session.__arrowrec
        }

        # configure credentials (if any) for organization
        if organization == 0:
            organization = self.PUBLIC_ORG
        self.authenticate(organization)

        # set cluster to desired number of nodes (if permitted based on credentials)
        self.scaleout(desired_nodes, time_to_live, bypass_dns)

    #
    #  source
    #
    def source (self, api, parm={}, stream=False, callbacks={}, path="/source", force_throw=False):
        '''
        handles making the HTTP request to the sliderule cluster nodes
        '''
        # initialize local variables
        rsps = {}
        headers = {'x-sliderule-client': f'python-{version.full_version}'}

        # build callbacks
        for c in self.callbacks:
            if c not in callbacks:
                callbacks[c] = self.callbacks[c]

        # Construct Request URL
        if self.service_org:
            url = 'https://%s.%s%s/%s' % (self.service_org, self.service_domain, path, api)
        else:
            url = 'http://%s%s/%s' % (self.service_domain, path, api)

        # Construct Payload
        if type(parm) == dict:
            payload = json.dumps(parm)
        else:
            payload = parm

        # Attempt request
        complete = False
        attempts = 3
        while not complete and attempts > 0:
            attempts -= 1
            try:
                # Build Authorization Header
                if self.service_org:
                    self.__buildauthheader(headers)

                # Perform Request
                if not stream:
                    data = self.session.get(url, data=payload, headers=headers, timeout=self.rqst_timeout, verify=self.ssl_verify)
                else:
                    data = self.session.post(url, data=payload, headers=headers, timeout=self.rqst_timeout, stream=True, verify=self.ssl_verify)
                data.raise_for_status()

                # Parse Response
                stream = self.__StreamSource(data)
                format = data.headers['Content-Type']
                if format == 'text/plain':
                    rsps = self.__parse_json(stream)
                elif format == 'application/json':
                    rsps = self.__parse_json(stream)
                elif format == 'application/octet-stream':
                    rsps = self.__parse_native(stream, callbacks)
                elif self.throw_exceptions or force_throw:
                    raise FatalError(f'Unsupported content type: {format}')
                else:
                    logger.error(f'Unsupported content type: {format}')

                # Success
                complete = True

            except requests.exceptions.SSLError as e:
                if self.throw_exceptions or force_throw:
                    raise FatalError(f'Exception in request to {url}: {e}')
                logger.error(f'Unable to verify SSL certificate for {url} ...retrying request')

            except requests.ConnectionError as e:
                if self.throw_exceptions or force_throw:
                    raise FatalError(f'Exception in request to {url}: {e}')
                logger.error(f'Connection error to endpoint {url} ...retrying request')

            except requests.Timeout as e:
                if self.throw_exceptions or force_throw:
                    raise FatalError(f'Exception in request to {url}: {e}')
                logger.error(f'Timed-out waiting for response from endpoint {url} ...retrying request')

            except requests.exceptions.ChunkedEncodingError as e:
                if self.throw_exceptions or force_throw:
                    raise FatalError(f'Exception in request to {url}: {e}')
                logger.error(f'Unexpected termination of response from endpoint {url} ...retrying request')

            except requests.HTTPError as e:
                if self.throw_exceptions or force_throw:
                    if e.response.status_code == 503:
                        raise FatalError(f'Server experiencing heavy load, stalling on request to {url}')
                    else:
                        raise FatalError(f'Exception in request to {url}: {e}')
                logger.error(f'HTTP error <{e.response.status_code}> returned in request to {url} ...retrying request')

        # Check Complete
        if not complete:
            if self.throw_exceptions or force_throw:
                raise FatalError(f'Request to {url} did not complete')
            else:
                rsps = None

        # Return Response
        return rsps

    #
    # get_definition
    #
    def get_definition (self, rectype, fieldname):
        '''
        returns the necessary parsing information for a field in a record
        '''
        # get record definition
        recdef = self.__populate(rectype)

        # return type information for field
        if fieldname in recdef and recdef[fieldname]["type"] in BASIC_TYPES:
            return BASIC_TYPES[recdef[fieldname]["type"]]
        else:
            return {}

    #
    #  set_verbose
    #
    def set_verbose (self, enable, loglevel=logging.INFO):
        '''
        set verbosity of log messages by adding/removing the console log
        handler and by setting the log level
        '''
        # massage loglevel parameter if passed in as a string
        if loglevel == "DEBUG":
            loglevel = logging.DEBUG
        elif loglevel == "INFO":
            loglevel = logging.INFO
        elif loglevel == "WARNING":
            loglevel = logging.WARNING
        elif loglevel == "WARN":
            loglevel = logging.WARN
        elif loglevel == "ERROR":
            loglevel = logging.ERROR
        elif loglevel == "FATAL":
            loglevel = logging.FATAL
        elif loglevel == "CRITICAL":
            loglevel = logging.CRITICAL

        # enable/disable logging to console
        if (enable == True) and (self.console == None):
            self.console = logging.StreamHandler()
            logger.addHandler(self.console)
        elif (enable == False) and (self.console != None):
            logger.removeHandler(self.console)
            self.console = None

        # always set level to requested
        logger.setLevel(loglevel)
        if self.console != None:
            self.console.setLevel(loglevel)

    #
    # update_available_servers
    #
    def update_available_servers (self, desired_nodes=None, time_to_live=None):
        '''
        makes a capacity request to the provisioning system (if requested) and
        returns the current capacity
        '''
        # initialize local variables
        requested_nodes = 0

        # update number of nodes
        if type(desired_nodes) == int:
            headers = {}
            rsps_body = {}
            requested_nodes = desired_nodes
            self.__buildauthheader(headers)

            # get boundaries of cluster and calculate nodes to request
            try:
                host = "https://ps." + self.service_domain + "/api/org_num_nodes/" + self.service_org + "/"
                rsps = self.session.get(host, headers=headers, timeout=self.rqst_timeout)
                rsps_body = rsps.json()
                rsps.raise_for_status()
                requested_nodes = max(min(desired_nodes, rsps_body["max_nodes"]), rsps_body["min_nodes"])
                if requested_nodes != desired_nodes:
                    logger.info("Provisioning system desired nodes overridden to {}".format(requested_nodes))
            except requests.exceptions.HTTPError as e:
                logger.info('{}'.format(e))
                logger.info('Provisioning system status request returned error => {}'.format(rsps_body["error_msg"]))
                if self.throw_exceptions:
                    raise

            # Request number of nodes in cluster
            try:
                if type(time_to_live) == int:
                    host = "https://ps." + self.service_domain + "/api/desired_org_num_nodes_ttl/" + self.service_org + "/" + str(requested_nodes) + "/" + str(time_to_live) + "/"
                    rsps = self.session.post(host, headers=headers, timeout=self.rqst_timeout)
                else:
                    host = "https://ps." + self.service_domain + "/api/desired_org_num_nodes/" + self.service_org + "/" + str(requested_nodes) + "/"
                    rsps = self.session.put(host, headers=headers, timeout=self.rqst_timeout)
                rsps_body = rsps.json()
                rsps.raise_for_status()
            except requests.exceptions.HTTPError as e:
                logger.info('{}'.format(e))
                logger.error('Provisioning system update request error => {}'.format(rsps_body["error_msg"]))
                if self.throw_exceptions:
                    raise

        # Get number of nodes currently registered
        try:
            rsps = self.source("status", parm={"service":"sliderule"}, path="/discovery", force_throw=True)
            available_servers = rsps["nodes"]
        except FatalError as e:
            logger.debug(f'Failed to retrieve number of nodes registered: {e}')
            available_servers = 0

        return available_servers, requested_nodes

    #
    # scaleout
    #
    def scaleout (self, desired_nodes, time_to_live, bypass_dns):
        '''
        makes a capacity request to the provisioning system and waits
        for the cluster to reach the requested capacity
        '''
        # check desired nodes
        if desired_nodes is None:
            return # nothing needs to be done
        if desired_nodes < 0:
            raise FatalError("Number of desired nodes must be greater than zero ({})".format(desired_nodes))

        # initialize DNS
        self.__initdns() # clear cache of DNS lookups for clusters
        if bypass_dns:
            self.__jamdns() # use ip address for cluster

        # send initial request for desired cluster state
        start = time.time()
        available_nodes,requested_nodes = self.update_available_servers(desired_nodes=desired_nodes, time_to_live=time_to_live)
        scale_up_needed = False

        # Wait for Cluster to Reach Desired State
        while available_nodes < requested_nodes:
            scale_up_needed = True
            logger.info("Waiting while cluster scales to desired capacity (currently at {} nodes, desired is {} nodes)... {} seconds".format(available_nodes, desired_nodes, int(time.time() - start)))
            time.sleep(10.0)
            available_nodes,_ = self.update_available_servers()
            # Override DNS if Cluster is Starting
            if available_nodes == 0 and not self.__dnsoverridden():
                self.__jamdns()
            # Timeout Occurred
            if int(time.time() - start) > self.MAX_PS_CLUSTER_WAIT_SECS:
                logger.error("Maximum time allowed waiting for cluster has been exceeded")
                break

        # Log Final Message if Cluster Needed State Change
        if scale_up_needed:
            logger.info("Cluster has reached capacity of {} nodes... {} seconds".format(available_nodes, int(time.time() - start)))

    #
    # authenticate
    #
    def authenticate (self, ps_organization, ps_username=None, ps_password=None, github_token=None):
        '''
        gets a bearer token (ps_access_token) from the provisioning system to use in requests
        to private clusters
        '''
        # initialize local variables
        login_status = False
        ps_url = "ps." + self.service_domain

        # set organization on any authentication request
        self.service_org = ps_organization

        # check for direct or public access
        if self.service_org == None:
            return True

        # attempt retrieving from environment
        if not github_token and not ps_username and not ps_password:
            github_token = os.environ.get("PS_GITHUB_TOKEN")
            ps_username = os.environ.get("PS_USERNAME")
            ps_password = os.environ.get("PS_PASSWORD")

        # attempt retrieving from netrc file
        if not github_token and not ps_username and not ps_password:
            try:
                netrc_file = netrc.netrc()
                login_credentials = netrc_file.hosts[ps_url]
                ps_username = login_credentials[0]
                ps_password = login_credentials[2]
            except Exception as e:
                pass

        # prompt user for username and password
        if not github_token and not ps_username and not ps_password:
            if ps_organization != self.PUBLIC_ORG:
                ps_username = input("Username: ")
                ps_password = getpass.getpass("Password: ")

        # build authentication request
        user = None
        if github_token:
            user = "github"
            rqst = {"org_name": ps_organization, "access_token": github_token}
            headers = {'Content-Type': 'application/json'}
            api = "https://" + ps_url + "/api/org_token_github/"
        elif ps_username or ps_password:
            user = "local"
            rqst = {"username": ps_username, "password": ps_password, "org_name": ps_organization}
            headers = {'Content-Type': 'application/json'}
            api = "https://" + ps_url + "/api/org_token/"

        # authenticate to provisioning system
        if user:
            try:
                rsps = self.session.post(api, data=json.dumps(rqst), headers=headers, timeout=self.rqst_timeout)
                rsps.raise_for_status()
                rsps = rsps.json()
                self.ps_refresh_token = rsps["refresh"]
                self.ps_access_token = rsps["access"]
                self.ps_token_exp =  time.time() + (float(rsps["access_lifetime"]) / 2)
                login_status = True
            except:
                if ps_organization != self.PUBLIC_ORG:
                    logger.error("Unable to authenticate %s user to %s" % (user, api))

        # return login status
        if ps_organization != self.PUBLIC_ORG:
            logger.info(f'Login status to {self.service_domain}/{self.service_org}: {login_status and "success" or "failure"}')
        return login_status

    #
    #  manager
    #
    def manager (self, api, content_json=True, as_post=False, headers={}):
        '''
        handles making the HTTP request to the sliderule manager
        '''
        # initialize local variables
        rsps = ""

        # Construct Request URL
        if self.service_org:
            url = 'https://%s.%s/manager/%s' % (self.service_org, self.service_domain, api)
        else:
            url = 'http://%s/manager/%s' % (self.service_domain, api)

        try:
            # Build Authorization Header
            if self.service_org:
                self.__buildauthheader(headers)

            # Perform Request
            if as_post:
                data = self.session.post(url, headers=headers, timeout=self.rqst_timeout, verify=self.ssl_verify)
            else:
                data = self.session.get(url, headers=headers, timeout=self.rqst_timeout, verify=self.ssl_verify)
            data.raise_for_status()

            # Parse Response
            stream = self.__StreamSource(data)
            lines = [line for line in stream]
            rsps = b''.join(lines)
            if content_json:
                rsps = json.loads(rsps)

        except requests.exceptions.SSLError as e:
            logger.error(f'Unable to verify SSL certificate for {url}')

        except requests.ConnectionError as e:
            logger.error(f'Connection error to endpoint {url}')

        except requests.Timeout as e:
            logger.error(f'Timed-out waiting for response from endpoint {url}')

        except requests.HTTPError as e:
            logger.error(f'HTTP error <{e.response.status_code}> returned in request to {url}: {rsps}')

        except Exception as e:
            logger.error(f'Failed to make request to {url}: {e}')

        # Return Response
        return rsps

    #
    # __StreamSource
    #
    class __StreamSource:
        def __init__(self, data):
            self.source = data
        def __iter__(self):
            for line in self.source.iter_content(None):
                yield line

    #
    # __BufferSource
    #
    class __BufferSource:
        def __init__(self, data):
            self.source = data
        def __iter__(self):
            yield self.source

    #
    #  __buildauthheader
    #
    def __buildauthheader(self, headers, force_refresh=False):
        '''
        builds the necessary authentication header for the http request to private
        clusters by using the bearer token from the provisioning system
        '''
        if self.ps_access_token:
            # Check if Refresh Needed
            with self.ps_lock:
                if time.time() > self.ps_token_exp or force_refresh:
                    host = "https://ps." + self.service_domain + "/api/org_token/refresh/"
                    rqst = {"refresh": self.ps_refresh_token}
                    hdrs = {'Content-Type': 'application/json', 'Authorization': 'Bearer ' + self.ps_access_token}
                    try:
                        rsps = self.session.post(host, data=json.dumps(rqst), headers=hdrs, timeout=self.rqst_timeout).json()
                        self.ps_refresh_token = rsps["refresh"]
                        self.ps_access_token = rsps["access"]
                        self.ps_token_exp =  time.time() + (float(rsps["access_lifetime"]) / 2)
                    except Exception as e:
                        logger.error(f'Failed to refresh token: {e}')
                        logger.error(f'Provisioning system response: {rsps}')
            # Build Authentication Header
            headers['Authorization'] = 'Bearer ' + self.ps_access_token

    #
    #  __populate
    #
    def __populate (self, rectype):
        '''
        rectype: record type supplied in response (string)
        '''
        if rectype not in self.recdef_table:
            self.recdef_table[rectype] = self.source("definition", {"rectype" : rectype})
        return self.recdef_table[rectype]

    #
    #  __parse_json
    #
    def __parse_json (self, data):
        """
        data: request response
        """
        lines = []
        for line in data:
            lines.append(line)
        response = b''.join(lines)
        return json.loads(response)

    #
    #  __decode_native
    #
    def __decode_native(self, rectype, rawdata):
        """
        rectype: record type supplied in response (string)
        rawdata: payload supplied in response (byte array)
        """
        # initialize record
        rec = { "__rectype": rectype }

        # get/populate record definition #
        recdef = self.__populate(rectype)

        # iterate through each field in definition
        for fieldname in recdef.keys():

            # double underline (__) as prefix indicates meta data
            if fieldname.find("__") == 0:
                continue

            # get field properties
            field = recdef[fieldname]
            ftype = field["type"]
            offset = int(field["offset"] / 8)
            elems = field["elements"]
            flags = field["flags"]

            # do not process pointers
            if "PTR" in flags:
                continue

            # check for mvp flag
            if not self.decode_aux and "AUX" in flags:
                continue

            # get endianness
            if "LE" in flags:
                endian = '<'
            else:
                endian = '>'

            # decode basic type
            if ftype in BASIC_TYPES:

                # check if array
                is_array = not (elems == 1)

                # get number of elements
                if elems <= 0:
                    elems = int((len(rawdata) - offset) / BASIC_TYPES[ftype]["size"])

                # build format string
                fmt = endian + str(elems) + BASIC_TYPES[ftype]["fmt"]

                # parse data
                value = struct.unpack_from(fmt, rawdata, offset)

                # set field
                if ftype == "STRING":
                    rec[fieldname] = ctypes.create_string_buffer(value[0]).value.decode('ascii')
                elif is_array:
                    rec[fieldname] = value
                else:
                    rec[fieldname] = value[0]

            # decode user type
            else:

                # populate record definition (if needed) #
                subrecdef = self.__populate(ftype)

                # check if array
                is_array = not (elems == 1)

                # get number of elements
                if elems <= 0:
                    elems = int((len(rawdata) - offset) / subrecdef["__datasize"])

                # return parsed data
                if is_array:
                    rec[fieldname] = []
                    for e in range(elems):
                        rec[fieldname].append(self.__decode_native(ftype, rawdata[offset:]))
                        offset += subrecdef["__datasize"]
                else:
                    rec[fieldname] = self.__decode_native(ftype, rawdata[offset:])

        # return record #
        return rec

    #
    #  __parse_native
    #
    def __parse_native (self, data, callbacks):
        """
        data: request response
        """
        recs = []

        rec_hdr_size = 8
        rec_size_index = 0
        rec_size_rsps = []

        rec_size = 0
        rec_index = 0
        rec_rsps = []

        for line in data:

            # Process Line Read
            i = 0
            while i < len(line):

                # Parse Record Size
                if(rec_size_index < rec_hdr_size):
                    bytes_available = len(line) - i
                    bytes_remaining = rec_hdr_size - rec_size_index
                    bytes_to_append = min(bytes_available, bytes_remaining)
                    rec_size_rsps.append(line[i:i+bytes_to_append])
                    rec_size_index += bytes_to_append
                    if(rec_size_index >= rec_hdr_size):
                        raw = b''.join(rec_size_rsps)
                        rec_version, rec_type_size, rec_data_size = struct.unpack('>hhi', raw)
                        if rec_version != 2:
                            raise FatalError("Invalid record format: %d" % (rec_version))
                        rec_size = rec_type_size + rec_data_size
                        rec_size_rsps.clear()
                    i += bytes_to_append

                # Parse Record
                elif(rec_size > 0):
                    bytes_available = len(line) - i
                    bytes_remaining = rec_size - rec_index
                    bytes_to_append = min(bytes_available, bytes_remaining)
                    rec_rsps.append(line[i:i+bytes_to_append])
                    rec_index += bytes_to_append
                    if(rec_index >= rec_size):
                        # Decode Record
                        rawbits = b''.join(rec_rsps)
                        rectype = ctypes.create_string_buffer(rawbits).value.decode('ascii')
                        rawdata = rawbits[len(rectype) + 1:]
                        rec     = self.__decode_native(rectype, rawdata)
                        if rectype == "conrec":
                            # parse records contained in record
                            buffer = self.__BufferSource(rawdata[rec["start"]:])
                            contained_recs = self.__parse_native(buffer, callbacks)
                            recs += contained_recs
                        else:
                            if callbacks != None and rectype in callbacks:
                                # Execute Call-Back on Record
                                callbacks[rectype](rec, self)
                            else:
                                # Append Record
                                recs.append(rec)
                        # Reset Record Parsing
                        rec_rsps.clear()
                        rec_size_index = 0
                        rec_size = 0
                        rec_index = 0
                    i += bytes_to_append

                # Zero Sized Record
                else:
                    rec_size_index = 0
                    rec_index = 0

        return recs

    #
    # __initdns
    #
    def __initdns (self):
        '''
        resets the global dictionary
        '''
        self.local_dns.clear()

    #
    # __dnsoverridden
    #
    def __dnsoverridden (self):
        '''
        checks if url is already overridden
        '''
        url = self.service_org + "." + self.service_domain
        return url.lower() in self.local_dns

    #
    # __jamdns
    #
    def __jamdns (self):
        '''
        override the dns entry
        '''
        global sliderule_dns
        headers = {}
        url = self.service_org + "." + self.service_domain
        self.__buildauthheader(headers)
        host = "https://ps." + self.service_domain + "/api/org_ip_adr/" + self.service_org + "/"
        rsps = self.session.get(host, headers=headers, timeout=self.rqst_timeout).json()
        if rsps["status"] == "SUCCESS":
            ipaddr = rsps["ip_address"]
            self.local_dns[url.lower()] = ipaddr
            logger.debug("Overriding DNS for {} with {}".format(url, ipaddr))
        sliderule_dns = self.local_dns

    #
    #  __logrec
    #
    @staticmethod
    def __logrec (rec, session):
        eventlogger[rec['level']]('%d:%s: %s' % (rec["time"], rec['source'], rec["message"]))

    #
    #  __alertrec
    #
    @staticmethod
    def __alertrec (rec, session):
        if rec["code"] == EXCEPTION_CODES["SIMPLIFY"]:
            raise RetryRequest("cmr simplification requested")
        elif rec["code"] < 0:
            eventlogger[rec["level"]]("Alert <%d>: %s", rec["code"], rec["text"])
        else:
            eventlogger[rec["level"]]("%s", rec["text"])

    #
    #  _arrowrec
    #
    @staticmethod
    def __arrowrec (rec, session):
        try :
            filename = rec["filename"]
            if rec["__rectype"] == 'arrowrec.meta':
                if filename in session.arrow_file_table:
                    raise FatalError(f'File transfer already in progress for {filename}')
                logger.info(f'Writing arrow file: {filename}')
                session.arrow_file_table[filename] = { "fp": open(filename, "wb"), "size": rec["size"], "progress": 0 }
            elif rec["__rectype"] == 'arrowrec.eof':
                logger.info(f'Checksum of arrow file: {rec["checksum"]}')
            else: # rec["__rectype"] == 'arrowrec.data'
                data = rec['data']
                file = session.arrow_file_table[filename]
                file["fp"].write(bytearray(data))
                file["progress"] += len(data)
                if file["progress"] >= file["size"]:
                    file["fp"].close()
                    logger.info(f'Closing arrow file: {filename}')
                    del session.arrow_file_table[filename]
        except Exception as e:
            raise FatalError(f'Failed to process arrow file: {e}')



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time
import numpy
import sliderule
from sliderule import logger

###############################################################################
# GLOBALS
###############################################################################

DATATYPES = {
    "TEXT":     0,
    "REAL":     1,
    "INTEGER":  2,
    "DYNAMIC":  3
}

ALL_ROWS = -1


###############################################################################
# APIs
###############################################################################

#
#  H5
#
def h5 (dataset, resource, asset, datatype=DATATYPES["DYNAMIC"], col=0, startrow=0, numrows=ALL_ROWS):
    '''
    Reads a dataset from an HDF5 file and returns the values of the dataset in a list

    This function provides an easy way for locally run scripts to get direct access to HDF5 data stored in a cloud environment.
    But it should be noted that this method is not the most efficient way to access remote H5 data, as the data is accessed one dataset at a time.
    The ``h5p`` api is the preferred solution for reading multiple datasets.

    One of the difficulties in reading HDF5 data directly from a Python script is converting the format of the data as it is stored in HDF5 to a data
    format that is easy to use in Python.  The compromise that this function takes is that it allows the user to supply the desired data type of the
    returned data via the **datatype** parameter, and the function will then return a **numpy** array of values with that data type.

    The data type is supplied as a ``DATATYPES`` enumeration:

    - ``DATATYPES["TEXT"]``: return the data as a string of unconverted bytes
    - ``DATATYPES["INTEGER"]``: return the data as an array of integers
    - ``DATATYPES["REAL"]``: return the data as an array of double precision floating point numbers
    - ``DATATYPES["DYNAMIC"]``: return the data in the numpy data type that is the closest match to the data as it is stored in the HDF5 file

    Parameters
    ----------
        dataset:    str
                    full path to dataset variable (e.g. ``/gt1r/geolocation/segment_ph_cnt``)
        resource:   str
                    HDF5 filename
        asset:      str
                    data source asset
        datatype:   int
                    the type of data the returned dataset list should be in (datasets that are naturally of a different type undergo a best effort conversion to the specified data type before being returned)
        col:        int
                    the column to read from the dataset for a multi-dimensional dataset; if there are more than two dimensions, all remaining dimensions are flattened out when returned.
        startrow:   int
                    the first row to start reading from in a multi-dimensional dataset (or starting element if there is only one dimension)
        numrows:    int
                    the number of rows to read when reading from a multi-dimensional dataset (or number of elements if there is only one dimension); if **ALL_ROWS** selected, it will read from the **startrow** to the end of the dataset.

    Returns
    -------
    numpy array
        dataset values

    Examples
    --------
        >>> segments    = h5.h5("/gt1r/land_ice_segments/segment_id",  resource, asset)
        >>> heights     = h5.h5("/gt1r/land_ice_segments/h_li",        resource, asset)
        >>> latitudes   = h5.h5("/gt1r/land_ice_segments/latitude",    resource, asset)
        >>> longitudes  = h5.h5("/gt1r/land_ice_segments/longitude",   resource, asset)
        >>> df = pd.DataFrame(data=list(zip(heights, latitudes, longitudes)), index=segments, columns=["h_mean", "latitude", "longitude"])
    '''
    datasets = [ { "dataset": dataset, "datatype": datatype, "col": col, "startrow": startrow, "numrows": numrows } ]
    values = h5p(datasets, resource, asset=asset)
    if len(values) > 0:
        return values[dataset]
    else:
        return numpy.empty(0)

#
#  Parallel H5
#
def h5p (datasets, resource, asset):
    '''
    Reads a list of datasets from an HDF5 file and returns the values of the dataset in a dictionary of lists.

    This function is considerably faster than the ``icesat2.h5`` function in that it not only reads the datasets in
    parallel on the server side, but also shares a file context between the reads so that portions of the file that
    need to be read multiple times do not result in multiple requests to S3.

    For a full discussion of the data type conversion options, see `h5 </web/rtd/api_reference/h5.html#h5>`_.

    Parameters
    ----------
        datasets:   dict
                    list of full paths to dataset variable (e.g. ``/gt1r/geolocation/segment_ph_cnt``); see below for additional parameters that can be added to each dataset
        resource:   str
                    HDF5 filename
        asset:      str
                    data source asset

    Returns
    -------
    dict
        numpy arrays of dataset values, where the keys are the dataset names

        The `datasets` dictionary can optionally contain the following elements per entry:

        * "valtype" (int): the type of data the returned dataset list should be in (datasets that are naturally of a different type undergo a best effort conversion to the specified data type before being returned)
        * "col" (int): the column to read from the dataset for a multi-dimensional dataset; if there are more than two dimensions, all remaining dimensions are flattened out when returned.
        * "startrow" (int): the first row to start reading from in a multi-dimensional dataset (or starting element if there is only one dimension)
        * "numrows" (int): the number of rows to read when reading from a multi-dimensional dataset (or number of elements if there is only one dimension); if **ALL_ROWS** selected, it will read from the **startrow** to the end of the dataset.

    Examples
    --------
        >>> from sliderule import icesat2
        >>> icesat2.init(["127.0.0.1"], False)
        >>> datasets = [
        ...         {"dataset": "/gt1l/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt1r/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt2l/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt2r/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt3l/land_ice_segments/h_li", "numrows": 5},
        ...         {"dataset": "/gt3r/land_ice_segments/h_li", "numrows": 5}
        ...     ]
        >>> rsps = h5.h5p(datasets, "ATL06_20181019065445_03150111_003_01.h5", "atlas-local")
        >>> print(rsps)
        {'/gt2r/land_ice_segments/h_li': array([45.3146427 , 45.27640582, 45.23608027, 45.21131015, 45.15692304]),
         '/gt2l/land_ice_segments/h_li': array([45.35118977, 45.33535027, 45.27195617, 45.21816889, 45.18534204]),
         '/gt1l/land_ice_segments/h_li': array([45.68811156, 45.71368944, 45.74234326, 45.74614113, 45.79866465]),
         '/gt3l/land_ice_segments/h_li': array([45.29602321, 45.34764226, 45.31430979, 45.31471701, 45.30034622]),
         '/gt1r/land_ice_segments/h_li': array([45.72632446, 45.76512574, 45.76337375, 45.77102473, 45.81307948]),
         '/gt3r/land_ice_segments/h_li': array([45.14954134, 45.18970635, 45.16637644, 45.15235916, 45.17135806])}
    '''
    # Baseline Request
    rqst = {
        "asset" : asset,
        "resource": resource,
        "datasets": datasets,
    }

    # Read H5 File
    try:
        rsps = sliderule.source("h5p", rqst, stream=True)
    except RuntimeError as e:
        logger.critical(e)
        rsps = []

    # Build Record Data
    results = {}
    for result in rsps:
        results[result["dataset"]] = sliderule.getvalues(result["data"], result["datatype"], result["size"])

    # Return Results
    return results



#!/usr/bin/env pythonw
""" Creates a cloud-optimized version of an ICESat-2 HDF5 file. """

# Copyright (c) 2024, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import sys
import argparse
import traceback
import h5py
from h5py import h5p, h5f
import numpy as np
import time

__version_= '1.0.1'

#-----------------------------------------------------------------------------
def h5_s2b(s:str) -> bytes:
  """ Converts a Python3 string to ASCII (bytes) for low-level calls.

  Args:
    s: Python string

  Returns:
    `s` as bytes
  """
  try:
    b = s.encode('ascii', 'ignore')
  except (UnicodeEncodeError, AttributeError):
    b = s
  return b

#------------------------------------------------------------------------------
def init():
  """ Parses arguments """

  parser = argparse.ArgumentParser(description="""Re-writes ICESat-2 HDF5 files using cloud optimization techniques.""")

  parser.add_argument('infile', help='Input HDF5 file.')

  parser.add_argument('outfile', help='Output HDF5 file.')

  parser.add_argument('-min_chunk', '--min_chunk', required=False,
                      type=int, default=500, help="""Do not resize any dataset with chunksize < this size (num elements).""")

  parser.add_argument('-seg_chunk', '--seg_chunk', required=False,
                      type=int, default=100000, help="""New chunksize for segment (non_photon) datasets (num elements).""")

  parser.add_argument('-ph_chunk', '--ph_chunk', required=False,
                      type=int, default=1000000, help="""New chunksize for any photon datasets (num elements).""")

  parser.add_argument('-gzip', '--gzip', required=False,
                      type=int, default=-1, help="""New gzip level (if dataset is compressed; 0=disable, -1=do not change)""")

  parser.add_argument('-fs_space_page', '--fs_space_page', dest='fs_space_page',
                      action='store_true', required=False, default=False,
                      help='Use H5F_FSPACE_STRATEGY_PAGE.')

  parser.add_argument('-fs_page_size', '--fs_page_size', required=False,
                      type=int, default=0., help="""Filespace page size (bytes), 0=do not change (requires -fs_space_page).""")

  parser.add_argument('-cache_slots', '--cache_slots',required=False,
                      type=int, default=0, help="""Number of cache slots, 0=do not change. (requires -fs_space_page)""")

  parser.add_argument('-chunk_cache', '--chunk_cache',required=False,
                      type=int, default=0., help="""Size of the chunk cache (bytes), 0=do not change. (requires -fs_space_page)""")

  parser.add_argument('-user_blocksize', '--user_blocksize', required=False,
                      type=int, default=0., help="""User block size (bytes), 0=do not change.""")

  parser.add_argument('-meta_blocksize', '--meta_blocksize', required=False,
                      type=int, default=0., help="""Metadata block size (bytes), 0=do not change.""")

  return parser.parse_args()

#------------------------------------------------------------------------------
def full_path(obj):
  """ Returns the full pathname of an object """

  if obj.parent is not None and obj.name != '/':
    return os.path.join(full_path(obj.parent), obj.name)
  else:
    return ''

#------------------------------------------------------------------------------
def print_stats_hdr():
  """ Prints the stats header """

  print('  {0:40s} {1:7s} {2:10s} {3:10s} {4:7s} {5:10s} {6:10s}'.format(
    'name',
    'src_zip',
    'src_size',
    'src_chunks',
    'dst_zip',
    'dst_nbytes',
    'dst_chunks',))
  print('  {0:40s} {1:7s} {2:10s} {3:10s} {4:7s} {5:10s} {6:10s}'.format(
    '----',
    '-------',
    '----------',
    '----------',
    '-------',
    '----------',
    '----------',))

#------------------------------------------------------------------------------
def print_stats(stats:dict):
  """ Prints stats """

  print('  {0:40s} {1:7d} {2:10d} {3:10d} {4:7d} {5:10d} {6:10d}'.format(
    stats['name'],
    stats['src_gzip'],
    stats['src_size'],
    stats['src_chunks'],
    stats['dst_gzip'],
    stats['dst_nbytes'],
    stats['dst_chunks'],))

#------------------------------------------------------------------------------
def copy_attributes(src, dest):
  """ Copies attributes from an input object to an output object """

  for key in src.attrs.keys():
    dest.attrs.create(key, data=src.attrs[key], dtype=src.attrs[key].dtype)

#------------------------------------------------------------------------------
def copy_dataset(args, src, dst, name, photons=False):
  """ Copies a dataset from an input group to an output group """

  stats = {
    'name':name,
    'src_gzip':0,
    'src_size':0,
    'src_nbytes':0,
    'src_chunks':0,
    'dst_gzip':0,
    'dst_nbytes':0,
    'dst_chunks':0,
    }

  src_dset = src[name]
  chunks = src_dset.chunks
  maxshape = src_dset.maxshape
  compress = src_dset.compression
  compress_opts = src_dset.compression_opts

  if src_dset.compression_opts is None:
    stats['src_gzip'] = -1
  else:
    stats['src_gzip'] = src_dset.compression_opts

  stats['src_size'] = src_dset.size
  stats['src_nbytes'] = src_dset.size

  if src_dset.chunks is None:
    stats['src_chunks'] = -1
  else:
    stats['src_chunks'] = src_dset.chunks[0]

  # Optionally modify compression (for compressed datasets)
  if compress == 'gzip':
    if args.gzip == 0:
      compress = None
      compress_opts = None
    elif args.gzip > 0:
      compress_opts = args.gzip

  # Optionally modify chunking (for select chunked datasets)
  #
  # This simplistic approach is probably not optimal for 2D or 3D
  # datasets, but it's a start.
  #
  # The 'variable' dimensions for ICESat-2 data are the left-most.
  # Do not change any chunksize less than args and do not change
  # any chunksize that is equal to maxshape.
  if maxshape[0] is None and chunks[0] is not None:
    if chunks[0] > args.min_chunk:
      chunks = list(chunks)
      if photons:
        chunks[0] = args.ph_chunk
      else:
        chunks[0] = args.seg_chunk
      chunks = tuple(chunks)

  dst_dset = dst.create_dataset(name, shape=src_dset.shape,
                                dtype=src_dset.dtype,
                                maxshape=src_dset.maxshape,
                                chunks=chunks,
                                compression=compress,
                                compression_opts=compress_opts,
                                shuffle=src_dset.shuffle,
                                fillvalue=src_dset.fillvalue,
                                track_order=True)
  dst_dset.write_direct(np.array(src_dset))
  if src_dset.is_scale:
    dst_dset.make_scale()

  copy_attributes(src_dset, dst_dset)

  if dst_dset.compression_opts is None:
    stats['dst_gzip'] = -1
  else:
    stats['dst_gzip'] = dst_dset.compression_opts

  stats['dst_size'] = dst_dset.size
  stats['dst_nbytes'] = dst_dset.size

  if dst_dset.chunks is None:
    stats['dst_chunks'] = -1
  else:
    stats['dst_chunks'] = dst_dset.chunks[0]

  print_stats(stats)

  return stats

#------------------------------------------------------------------------------
def copy_group(args, src, dst, name):
  """ Copies a group from an input file to an output file

  Could do something with stats here.  ie:
    min/max within each group, etc.  Not sure its worth it now.
  """

  src_grp = src[name]
  if name != '/':
    dst_grp = dst.create_group(name, track_order=True)
  else:
    dst_grp = dst

  # Copy group attributes
  copy_attributes(src_grp, dst_grp)

  path = full_path(src_grp)
  print(f'\nGroup: {path}\n')
  print_stats_hdr()

  # This hack will only work for ATL03
  if name == 'heights':
    photons = True
  else:
    photons = False

  # Copy child datasets
  for item in src_grp.keys():
    if isinstance(src_grp[item], h5py.Dataset):
      if item == 'control':
        continue
      stats = copy_dataset(args, src_grp, dst_grp, item, photons=photons)

  # Copy child groups
  for item in src_grp.keys():
    if isinstance(src_grp[item], h5py.Group):
      copy_group(args, src_grp, dst_grp, item)

#------------------------------------------------------------------------------
def optimize_file(args):
  """ Copies an ICESat-2 HDF5 file to a new cloud-optimized file.

  Uses the low-level h5py APID to open/query the input file and to create
  the output file. I started using existing code that was writted with an
  old h5py and then realized (too late) new h5py supports all these things.

  So, then switches to the high-level APID to actually copy the data.
  """

  strategies = [
    'H5F_FSPACE_STRATEGY_FSM_AGGR',
    'H5F_FSPACE_STRATEGY_PAGE',
    'H5F_FSPACE_STRATEGY_AGGR',
    'H5F_FSPACE_STRATEGY_NONE',
   ]

  if not os.path.isfile(args.infile):
    print(f'{args.infile} : file does not exist.')
    sys.exit(3)

  # Open the input file
  try:
    in_fid = h5f.open(h5_s2b(args.infile), h5f.ACC_RDONLY)
  except Exception as e:
    print(f'Error opening input file - {e}')
    sys.exit(3)

  # Get info from the file creation plist
  try:
    fcpl = in_fid.get_create_plist()
    fs_strategy = fcpl.get_file_space_strategy()
    fspace_page_size = fcpl.get_file_space_page_size()
    user_blocksize = fcpl.get_userblock()
  except Exception as e:
    print(f'Error reading the file creation plist - {e}')
    in_fid.close()
    sys.exit(3)

  # Get info from the file access plist
  try:
    fapl = in_fid.get_access_plist()
    cache_info = fapl.get_cache()
    meta_blocksize = fapl.get_meta_block_size()
  except Exception as e:
    print(f'Error reading the file access plist - {e}')
    in_fid.close()
    sys.exit(3)

  cache_slots = cache_info[1]
  chunk_cache = cache_info[2]

  strategy_str = strategies[fs_strategy[0]]
  print('--')
  print(f'Original fs_strategy      = {strategy_str}')
  print(f'Original fspace_page_size = {fspace_page_size}')
  print(f'Original cache_slots      = {cache_slots}')
  print(f'Original chunk_cache      = {chunk_cache}')
  print(f'Original userblock_size   = {user_blocksize}')
  print(f'Original meta_blocksize   = {meta_blocksize}')
  print('--')

  # Apply any optional argument values
  if args.fs_space_page:
    fs_strategy = list(fs_strategy)
    cache_info = list(cache_info)
    fs_strategy[0] = 1
    if args.fs_page_size > 0.:
      fspace_page_size = args.fs_page_size
    if args.chunk_cache > 0.:
      cache_info[2] = args.chunk_cache
    if args.cache_slots > 0.:
      cache_info[1] = args.cache_slots
    fs_strategy = tuple(fs_strategy)
    cache_info = tuple(cache_info)

  if args.user_blocksize > 0.:
    user_blocksize = args.user_blocksize
  if args.meta_blocksize > 0.:
    meta_blocksize = args.meta_blocksize

  # Create a new file create property list
  try:
    fcpl = h5p.create(h5p.FILE_CREATE)
    fcpl.set_link_creation_order(h5p.CRT_ORDER_TRACKED|h5p.CRT_ORDER_INDEXED)
    fcpl.set_attr_creation_order(h5p.CRT_ORDER_TRACKED|h5p.CRT_ORDER_INDEXED)
    fcpl.set_file_space_strategy(fs_strategy[0], fs_strategy[1], fs_strategy[2])
    fcpl.set_file_space_page_size(fspace_page_size)
    fcpl.set_userblock(user_blocksize)
  except Exception as e:
    print(f'Error creating the file creation plist - {e}')
    in_fid.close()
    sys.exit(3)

  # Create a new file access property list
  try:
    fapl = h5p.create(h5p.FILE_ACCESS)
    fapl.set_cache(cache_info[0], cache_info[1], cache_info[2], cache_info[3])
    fapl.set_meta_block_size(meta_blocksize)
  except Exception as e:
    print(f'Error creating the file access plist - {e}')
    in_fid.close()
    sys.exit(3)

  # Create the output file
  try:
    out_fid = h5f.create(h5_s2b(args.outfile), h5f.ACC_TRUNC, fcpl=fcpl, fapl=fapl)
  except Exception as e:
    print(f'Error creating the output file - {e}')
    sys.exit(3)

  # Get info from the file creation plist
  try:
    fcpl = out_fid.get_create_plist()
    fs_strategy = fcpl.get_file_space_strategy()
    fspace_page_size = fcpl.get_file_space_page_size()
    user_blocksize = fcpl.get_userblock()
  except Exception as e:
    print(f'Error reading the file creation plist - {e}')
    in_fid.close()
    out_fid.close()
    sys.exit(3)

  # Get info from the file access plist
  try:
    fapl = out_fid.get_access_plist()
    cache_info = fapl.get_cache()
    meta_blocksize = fapl.get_meta_block_size()
  except Exception as e:
    print(f'Error reading the file access plist - {e}')
    in_fid.close()
    out_fid.close()
    sys.exit(3)

  cache_slots = cache_info[1]
  chunk_cache = cache_info[2]

  strategy_str = strategies[fs_strategy[0]]
  print('--')
  print(f'Revised fs_strategy       = {strategy_str}')
  print(f'Revised fspace_page_size  = {fspace_page_size}')
  print(f'Revised cache_slots       = {cache_slots}')
  print(f'Revised chunk_cache       = {chunk_cache}')
  print(f'Revised userblock_size    = {user_blocksize}')
  print(f'Revised meta_blocksize    = {meta_blocksize}')
  print('--')

  # Close the files so we can reopen them in the high-level APID
  in_fid.close()
  out_fid.close()

  # Read from input and write to output
  try:

    beg_time = time.time()
    with h5py.File(args.infile, 'r') as src, \
         h5py.File(args.outfile, 'r+') as dst:
      copy_group(args, src, dst, '/')
    end_time = time.time()

  except Exception as e:
    print(f'Error reading/writing data - {e}')
    print(traceback.format_exc())
    in_fid.close()
    out_fid.close()
    sys.exit(3)

  # Print file sizes/change
  src_stats = os.stat(args.infile)
  dst_stats = os.stat(args.outfile)
  src_mb = src_stats.st_size / (1024 * 1024)
  dst_mb = dst_stats.st_size / (1024 * 1024)
  schange = (dst_mb / src_mb) * 100.
  print('--')
  print('Total read/write time=', end_time - beg_time)
  print(f'File Sizes (MiB): input={src_mb}, output={dst_mb}, change={schange}%')
  print('--')

#------------------------------------------------------------------------------
def run():

  args = init()
  optimize_file(args)

#-------------------------------------------------------------------------------
if __name__ == '__main__':
  run()



# Copyright (c) 2021, University of Washington
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# 3. Neither the name of the University of Washington nor the names of its
#    contributors may be used to endorse or promote products derived from this
#    software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE UNIVERSITY OF WASHINGTON AND CONTRIBUTORS
# “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE UNIVERSITY OF WASHINGTON OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
# OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import numpy
import geopandas
import sliderule
from sliderule import logger
from sliderule.session import Session

###############################################################################
# GLOBAL DATA
###############################################################################

# CRS used in GEDI Standard Data Products
GEDI_CRS = "EPSG:7912"

###############################################################################
# LOCAL FUNCTIONS
###############################################################################

#
# Flatten Batches
#
def __flattenbatches(rsps, rectype, batch_column, parm, keep_id, as_numpy_array, height_key):

    # Check Responses
    if rsps == None:
        return sliderule.emptyframe(crs=GEDI_CRS)

    # Check for Output Options
    if "output" in parm:
        gdf = sliderule.procoutputfile(parm, rsps)
        return gdf

    # Flatten Records
    columns = {}
    records = []
    num_records = 0
    field_dictionary = {} # [<field_name>] = {"shot_number": [], <field_name>: []}
    file_dictionary = {} # [id] = "filename"
    if len(rsps) > 0:
        # Sort Records
        for rsp in rsps:
            if rectype in rsp['__rectype']:
                records += rsp,
                num_records += len(rsp[batch_column])
            elif 'ancfrec' == rsp['__rectype']:
                for field_rec in rsp['fields']:
                    shot_number = numpy.uint64(rsp['extent_id'])
                    field_name = parm['anc_fields'][field_rec['field_index']]
                    if field_name not in field_dictionary:
                        field_dictionary[field_name] = {'shot_number': [], field_name: []}
                    field_dictionary[field_name]['shot_number'] += shot_number,
                    field_dictionary[field_name][field_name] += sliderule.getvalues(field_rec['value'], field_rec['datatype'], len(field_rec['value']), num_elements=1)[0],
            elif 'ancerec' == rsp['__rectype']:
                shot_number = numpy.uint64(rsp['extent_id'])
                field_name = parm['anc_fields'][rsp['field_index']]
                if field_name not in field_dictionary:
                    field_dictionary[field_name] = {'shot_number': [], field_name: []}
                field_dictionary[field_name]['shot_number'] += shot_number,
                field_dictionary[field_name][field_name] += sliderule.getvalues(rsp['data'], rsp['datatype'], len(rsp['data']), num_elements=rsp['num_elements']),
            elif 'rsrec' == rsp['__rectype'] or 'zsrec' == rsp['__rectype']:
                if rsp["num_samples"] <= 0:
                    continue
                # Get field names and set
                sample = rsp["samples"][0]
                field_names = list(sample.keys())
                field_names.remove("__rectype")
                field_set = rsp['key']
                if rsp["num_samples"] > 1:
                    as_numpy_array = True
                # On first time, build empty dictionary for field set associated with raster
                if field_set not in field_dictionary:
                    field_dictionary[field_set] = {'shot_number': []}
                    for field in field_names:
                        field_dictionary[field_set][field_set + "." + field] = []
                # Populate dictionary for field set
                field_dictionary[field_set]['shot_number'] += numpy.uint64(rsp['index']),
                for field in field_names:
                    if as_numpy_array:
                        data = []
                        for s in rsp["samples"]:
                            data += s[field],
                        field_dictionary[field_set][field_set + "." + field] += numpy.array(data),
                    else:
                        field_dictionary[field_set][field_set + "." + field] += sample[field],
            elif 'fileidrec' == rsp['__rectype']:
                file_dictionary[rsp["file_id"]] = rsp["file_name"]

        # Build Columns
        if num_records > 0:
            # Initialize Columns
            sample_record = records[0][batch_column][0]
            for field in sample_record.keys():
                fielddef = sliderule.getdefinition(sample_record['__rectype'], field)
                if len(fielddef) > 0:
                    if type(sample_record[field]) == tuple:
                        columns[field] = numpy.empty(num_records, dtype=object)
                    else:
                        columns[field] = numpy.empty(num_records, fielddef["nptype"])
            # Populate Columns
            cnt = 0
            for record in records:
                for batch in record[batch_column]:
                    for field in columns:
                        columns[field][cnt] = batch[field]
                    cnt += 1
    else:
        logger.debug("No response returned")

    # Build Initial GeoDataFrame
    gdf = sliderule.todataframe(columns, height_key=height_key, crs=GEDI_CRS)

    # Merge Ancillary Fields
    for field_set in field_dictionary:
        df = geopandas.pd.DataFrame(field_dictionary[field_set])
        gdf = geopandas.pd.merge(gdf, df, how='left', on='shot_number').set_axis(gdf.index)

    # Delete Shot Number Column
    if len(gdf) > 0 and not keep_id:
        del gdf["shot_number"]

    # Attach Metadata
    if len(file_dictionary) > 0:
        gdf.attrs['file_directory'] = file_dictionary

    # Return GeoDataFrame
    return gdf

#
#  Perform Processing Request
#
def __processing_request(parm, asset, callbacks, resources, keep_id, as_numpy_array, api, rec, height_key):

    # Default the Asset
    rqst_parm = parm.copy()
    if "asset" not in rqst_parm:
        rqst_parm["asset"] = asset

    # Build GEDI Request
    rqst = {
        "resources": resources,
        "parms": rqst_parm
    }

    # Make API Processing Request
    rsps = sliderule.source(api, rqst, stream=True, callbacks=callbacks)

    # Return GeoDataFrame
    return __flattenbatches(rsps, rec, 'footprint', parm, keep_id, as_numpy_array, height_key)

###############################################################################
# APIs
###############################################################################

#
#  Initialize
#
def init (url=Session.PUBLIC_URL, verbose=False, loglevel=logging.CRITICAL, organization=Session.PUBLIC_ORG, desired_nodes=None, time_to_live=60, bypass_dns=False):
    '''
    Initializes the Python client for use with SlideRule and should be called before other GEDI API calls.
    This function is a wrapper for the `sliderule.init(...) function </web/rtds/api_reference/sliderule.html#init>`_.

    Examples
    --------
        >>> from sliderule import gedi
        >>> gedi.init()
    '''
    sliderule.init(url, verbose, loglevel, organization, desired_nodes, time_to_live, bypass_dns)

#
#  GEDI L4A
#
def gedi04a (parm, resource):
    '''
    Performs GEDI L4A subsetting of elevation footprints

    Parameters
    ----------
    parms:      dict
                parameters used to configure subsetting process
    resource:   str
                GEDI HDF5 filename
    asset:      str
                data source asset

    Returns
    -------
    GeoDataFrame
        gridded footrpints
    '''
    return gedi04ap(parm, resources=[resource])

#
#  Parallel GEDI04A
#
def gedi04ap(parm, callbacks={}, resources=None, keep_id=False, as_numpy_array=False, height_key=None):
    '''
    Performs subsetting in parallel on GEDI data and returns elevation footprints.  This function expects that the **parm** argument
    includes a polygon which is used to fetch all available resources from the CMR system automatically.  If **resources** is specified
    then any polygon or resource filtering options supplied in **parm** are ignored.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure subsetting process
        asset:          str
                        data source asset
        callbacks:      dictionary
                        a callback function that is called for each result record
        resources:      list
                        a list of granules to process (e.g. ["GEDI04_A_2019229131935_O03846_02_T03642_02_002_02_V002.h5", ...])
        keep_id:        bool
                        whether to retain the "extent_id" column in the GeoDataFrame for future merges
        as_numpy_array: bool
                        whether to provide all sampled values as numpy arrays even if there is only a single value
        height_key:     str
                        identifies the name of the column provided for the 3D CRS transformation

    Returns
    -------
    GeoDataFrame
        geolocated footprints

    Examples
    --------
        >>> from sliderule import gedi
        >>> gedi.init()
        >>> region = [ {"lon":-105.82971551223244, "lat": 39.81983728534918},
        ...            {"lon":-105.30742121965137, "lat": 39.81983728534918},
        ...            {"lon":-105.30742121965137, "lat": 40.164048017973755},
        ...            {"lon":-105.82971551223244, "lat": 40.164048017973755},
        ...            {"lon":-105.82971551223244, "lat": 39.81983728534918} ]
        >>> parms = { "poly": region }
        >>> resources = ["GEDI04_A_2019229131935_O03846_02_T03642_02_002_02_V002.h5"]
        >>> asset = "ornldaac-s3"
        >>> rsps = gedi.gedi04ap(parms, asset=asset, resources=resources)
    '''
    return __processing_request(parm, "gedil4a", callbacks, resources, keep_id, as_numpy_array, 'gedi04ap', 'gedi04arec', height_key)

#
#  GEDI L2A
#
def gedi02a (parm, resource):
    '''
    Performs GEDI L2A subsetting of elevation footprints

    Parameters
    ----------
    parms:      dict
                parameters used to configure subsetting process
    resource:   str
                GEDI HDF5 filename
    asset:      str
                data source asset

    Returns
    -------
    GeoDataFrame
        gridded footrpints
    '''
    return gedi02ap(parm, resources=[resource])

#
#  Parallel GEDI02A
#
def gedi02ap(parm, callbacks={}, resources=None, keep_id=False, as_numpy_array=False, height_key=None):
    '''
    Performs subsetting in parallel on GEDI data and returns geolocated footprints.  This function expects that the **parm** argument
    includes a polygon which is used to fetch all available resources from the CMR system automatically.  If **resources** is specified
    then any polygon or resource filtering options supplied in **parm** are ignored.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure subsetting process
        asset:          str
                        data source asset
        callbacks:      dictionary
                        a callback function that is called for each result record
        resources:      list
                        a list of granules to process (e.g. ["GEDI04_A_2019229131935_O03846_02_T03642_02_002_02_V002.h5", ...])
        keep_id:        bool
                        whether to retain the "extent_id" column in the GeoDataFrame for future merges
        as_numpy_array: bool
                        whether to provide all sampled values as numpy arrays even if there is only a single value
        height_key:     str
                        identifies the name of the column provided for the 3D CRS transformation

    Returns
    -------
    GeoDataFrame
        geolocated footprints

    Examples
    --------
        >>> from sliderule import gedi
        >>> gedi.init()
        >>> region = [ {"lon":-105.82971551223244, "lat": 39.81983728534918},
        ...            {"lon":-105.30742121965137, "lat": 39.81983728534918},
        ...            {"lon":-105.30742121965137, "lat": 40.164048017973755},
        ...            {"lon":-105.82971551223244, "lat": 40.164048017973755},
        ...            {"lon":-105.82971551223244, "lat": 39.81983728534918} ]
        >>> parms = { "poly": region }
        >>> resources = ["GEDI02_A_2019229131935_O03846_02_T03642_02_002_02_V002.h5"]
        >>> asset = "gedi-local"
        >>> rsps = gedi.gedi02ap(parms, asset=asset, resources=resources)
    '''
    return __processing_request(parm, "gedil2a", callbacks, resources, keep_id, as_numpy_array, 'gedi02ap', 'gedi02arec', height_key)

#
#  GEDI L1B
#
def gedi01b (parm, resource):
    '''
    Performs GEDI L1B subsetting of elevation waveforms

    Parameters
    ----------
    parms:      dict
                parameters used to configure subsetting process
    resource:   str
                GEDI HDF5 filename
    asset:      str
                data source asset

    Returns
    -------
    GeoDataFrame
        gridded footrpints
    '''
    return gedi01bp(parm, resources=[resource])

#
#  Parallel GEDI01B
#
def gedi01bp(parm, callbacks={}, resources=None, keep_id=False, as_numpy_array=False, height_key=None):
    '''
    Performs subsetting in parallel on GEDI data and returns geolocated footprints.  This function expects that the **parm** argument
    includes a polygon which is used to fetch all available resources from the CMR system automatically.  If **resources** is specified
    then any polygon or resource filtering options supplied in **parm** are ignored.

    Parameters
    ----------
        parms:          dict
                        parameters used to configure subsetting process
        asset:          str
                        data source asset
        callbacks:      dictionary
                        a callback function that is called for each result record
        resources:      list
                        a list of granules to process (e.g. ["GEDI04_A_2019229131935_O03846_02_T03642_02_002_02_V002.h5", ...])
        keep_id:        bool
                        whether to retain the "extent_id" column in the GeoDataFrame for future merges
        as_numpy_array: bool
                        whether to provide all sampled values as numpy arrays even if there is only a single value
        height_key:     str
                        identifies the name of the column provided for the 3D CRS transformation

    Returns
    -------
    GeoDataFrame
        geolocated footprints

    Examples
    --------
        >>> from sliderule import gedi
        >>> gedi.init()
        >>> region = [ {"lon":-105.82971551223244, "lat": 39.81983728534918},
        ...            {"lon":-105.30742121965137, "lat": 39.81983728534918},
        ...            {"lon":-105.30742121965137, "lat": 40.164048017973755},
        ...            {"lon":-105.82971551223244, "lat": 40.164048017973755},
        ...            {"lon":-105.82971551223244, "lat": 39.81983728534918} ]
        >>> parms = { "poly": region }
        >>> resources = ["GEDI01_B_2019229131935_O03846_02_T03642_02_002_02_V002.h5"]
        >>> asset = "gedi-local"
        >>> rsps = gedi.gedi01bp(parms, asset=asset, resources=resources)
    '''
    return __processing_request(parm, "gedil1b", callbacks, resources, keep_id, as_numpy_array, 'gedi01bp', 'gedi01brec', height_key)



from datetime import datetime as dt
import pandas as pd
import requests
import sys


# OpenDAP file structure:
# http://oceandata.sci.gsfc.nasa.gov/opendap/VIIRSJ1/L3SMI/2019/0101/JPSS1_VIIRS.20190101_20190108.L3m.8D.KD.Kd_490.4km.nc.dap.nc4

years = ['2018', '2019', '2020', '2021', '2022', '2023', '2024']

for year in years:
    # get start datetime from the granule, establish the 8-day cycle start dates for
    #   the year, generate the time deltas to the granule date
    eight_day = pd.date_range(start='1/1/%s'%year, end='12/31/%s'%year,freq='8d')
    for dl_start in eight_day:
        dl_end = dl_start + pd.Timedelta('7d')
        if dl_end.year != dl_start.year:
            dl_end = dt.strptime("%s-12-31"%dl_start.year,"%Y-%m-%d")

        # build the download string
        oc_start = 'http://oceandata.sci.gsfc.nasa.gov/opendap/VIIRSJ1/L3SMI/%s/%s/'%(dl_start.year,dl_start.strftime('%m%d'))
        oc_dates = 'JPSS1_VIIRS.%s_%s'%(dl_start.strftime('%Y%m%d'), dl_end.strftime('%Y%m%d'))
        oc_end = ".L3m.8D.KD.Kd_490.4km.nc"
        dap_end = '.dap.nc4'
        oc_file = oc_start + oc_dates + oc_end + dap_end

        print(oc_file)


        r = requests.get(oc_file)

        if r.ok:
            with open("test.out",'wb') as f:
            
                # write the contents of the response (r.content)
                # to a new file in binary mode.
                f.write(r.content)                
        sys.exit(0)
        